﻿using System;
using System.Diagnostics;
using System.ComponentModel;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using SlimDX;

//Some namespace mappings
using D2D = SlimDX.Direct2D;

namespace Articles.Tutorials.Direct2D
{
    public partial class Chapter2 : UserControl
    {
        //Resources for Direct2D rendering
        private D2D.Factory m_factory;
        private D2D.WindowRenderTarget m_renderTarget;
        private D2D.LinearGradientBrush m_backBrushEx;
        private D2D.GradientStopCollection m_backBrushGradient;
        private D2D.SolidColorBrush m_blackBrush;
        private D2D.SolidColorBrush m_lightBlueBrush;

        //Some generic members
        private Brush m_backBrushGdi;
        private Pen m_blackBrushGdi;
        private SolidBrush m_lightBlueBrushGdi;
        private bool m_initialized;
        private bool m_debugMode;

        /// <summary>
        /// Initializes a new instance of the <see cref="Direct2DPanel"/> class.
        /// </summary>
        public Chapter2()
        {
            this.InitializeComponent();

            //Update control styles
            this.SetStyle(ControlStyles.AllPaintingInWmPaint, true);
            this.SetStyle(ControlStyles.Opaque, true);
            this.SetStyle(ControlStyles.ResizeRedraw, true);

            //Update back brush
            m_backBrushGdi = new SolidBrush(this.BackColor);
        }

        /// <summary>
        /// Loads all graphics resources.
        /// </summary>
        private void InitializeGraphics()
        {
            //Get requested debug level
            D2D.DebugLevel debugLevel = D2D.DebugLevel.None;
            if (m_debugMode) { debugLevel = D2D.DebugLevel.Error; }

            //Create factory object
            m_factory = new D2D.Factory(D2D.FactoryType.SingleThreaded, debugLevel);

            //Create the render target
            m_renderTarget = new D2D.WindowRenderTarget(m_factory, new D2D.WindowRenderTargetProperties()
            {
                Handle = this.Handle,
                PixelSize = this.Size,
                PresentOptions = D2D.PresentOptions.Immediately
            });

            //Create linear gradient brush
            D2D.GradientStop[] gradientStops = new D2D.GradientStop[]
            {
                new D2D.GradientStop(){ Position = 0f, Color = new Color4(Color.LightGray) },
                new D2D.GradientStop(){ Position = 1f, Color = new Color4(Color.LightSteelBlue) }
            };
            m_backBrushGradient = new D2D.GradientStopCollection(m_renderTarget, gradientStops);
            m_backBrushEx = new D2D.LinearGradientBrush(
                m_renderTarget,
                m_backBrushGradient,
                new D2D.LinearGradientBrushProperties()
                {
                    StartPoint = new PointF(0, this.Height),
                    EndPoint = new PointF(0, 0)
                });

            //Create solid brush
            m_blackBrush = new D2D.SolidColorBrush(m_renderTarget, new Color4(Color.Black));
            m_lightBlueBrush = new D2D.SolidColorBrush(m_renderTarget, new Color4(Color.LightBlue));
            m_blackBrush.Opacity = 0.4f;
            m_lightBlueBrush.Opacity = 0.4f;

            //Create solid brush (gdi version)
            m_blackBrushGdi = new Pen(Color.FromArgb(100, Color.Black));
            m_lightBlueBrushGdi = new SolidBrush(Color.FromArgb(100, Color.LightBlue));

            //Create gdi backbrush
            if (m_backBrushGdi != null) { m_backBrushGdi.Dispose(); }
            m_backBrushGdi = new LinearGradientBrush(
                new Point(0, this.Height),
                new Point(0, 0),
                Color.LightGray,
                Color.LightSteelBlue);

            //Update initialization flag
            m_initialized = true;
        }

        /// <summary>
        /// Unloads all graphics resources.
        /// </summary>
        private void UnloadGraphics()
        {
            if (m_lightBlueBrush != null) { m_lightBlueBrush.Dispose(); }
            if (m_blackBrush != null) { m_blackBrush.Dispose(); }
            if (m_backBrushEx != null) { m_backBrushEx.Dispose(); }
            if (m_backBrushGradient != null) { m_backBrushGradient.Dispose(); }
            if (m_renderTarget != null) { m_renderTarget.Dispose(); }
            if (m_factory != null) { m_factory.Dispose(); }

            m_lightBlueBrush = null;
            m_backBrushEx = null;
            m_backBrushGradient = null;
            m_renderTarget = null;
            m_factory = null;
        }

        /// <summary>
        /// Called when window opens.
        /// </summary>
        /// <param name="e">An <see cref="T:System.EventArgs"/> that contains the event data.</param>
        protected override void OnHandleCreated(EventArgs e)
        {
            base.OnHandleCreated(e);

            if ((!m_initialized) || (!this.DesignMode))
            {
                InitializeGraphics();
            }
        }

        /// <summary>
        /// Called when window closes.
        /// </summary>
        /// <param name="e">An <see cref="T:System.EventArgs"/> that contains the event data.</param>
        protected override void OnHandleDestroyed(EventArgs e)
        {
            base.OnHandleDestroyed(e);

            if (m_initialized)
            {
                UnloadGraphics();
            }
        }

        /// <summary>
        /// Called when size of the control changed.
        /// </summary>
        /// <param name="e">An <see cref="T:System.EventArgs"/> that contains the event data.</param>
        protected override void OnResize(EventArgs e)
        {
            base.OnResize(e);

            if ((this.Width > 0) && (this.Height > 0))
            {
                if (m_backBrushGdi != null) { m_backBrushGdi.Dispose(); }
                m_backBrushGdi = new LinearGradientBrush(
                    new Point(0, this.Height),
                    new Point(0, 0),
                    Color.LightGray,
                    Color.LightSteelBlue);

                if (m_initialized)
                {
                    //Resize render target
                    m_renderTarget.Resize(this.Size);
                    m_backBrushEx.StartPoint = new PointF(0, this.Height);
                }
            }
        }

        /// <summary>
        /// User changes the size of the rectangle.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="System.Windows.Forms.ScrollEventArgs"/> instance containing the event data.</param>
        private void OnScrollRectangleSizeScroll(object sender, ScrollEventArgs e)
        {
            this.Invalidate();
        }

        /// <summary>
        /// Called when user changes a parameter.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void OnChkParameterCheckedChanged(object sender, EventArgs e)
        {
            this.Invalidate();
        }

        /// <summary>
        /// Called when user changes the renderer.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void OnRadGdiCheckedChanged(object sender, EventArgs e)
        {
            if (m_radGdi.Checked)
            {
                m_chkRounded.Enabled = false;
                m_chkRounded.Checked = false;
                this.DoubleBuffered = true;
                this.Invalidate();
            }
        }

        /// <summary>
        /// Called when user changes the renderer.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void OnRadD2dCheckedChanged(object sender, EventArgs e)
        {
            if (m_radD2d.Checked)
            {
                m_chkRounded.Enabled = true;
                this.DoubleBuffered = false;
                this.Invalidate();
            }
        }

        /// <summary>
        /// Called when control has to paint itself.
        /// </summary>
        /// <param name="e">A <see cref="T:System.Windows.Forms.PaintEventArgs"/> that contains the event data.</param>
        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);

            if (m_initialized && m_radD2d.Checked)
            {
                m_renderTarget.BeginDraw();
                try
                {
                if (m_chkHighQuality.Checked) { m_renderTarget.AntialiasMode = D2D.AntialiasMode.PerPrimitive; }
                else { m_renderTarget.AntialiasMode = D2D.AntialiasMode.Aliased; }

                //Fill background
                m_renderTarget.FillRectangle(
                    m_backBrushEx,
                    new Rectangle(new Point(), this.Size));

                //Update transparency of brushes
                if (m_chkTransparent.Checked)
                {
                    m_blackBrush.Opacity = 0.5f;
                    m_lightBlueBrush.Opacity = 0.5f;
                }
                else
                {
                    m_blackBrush.Opacity = 1f;
                    m_lightBlueBrush.Opacity = 1f;
                }

                //Draw all rectangles
                int rectSize = m_scrollRectangleSize.Value;
                int rectSpacing = m_scrollRectangleSize.Value / 10;
                int radius = m_scrollRectangleSize.Value / 7;
                int rectangleCount = 0;
                Stopwatch stopwatch = new Stopwatch();
                stopwatch.Start();
                for (int loopX = 0; loopX < this.Width / rectSize + 1; loopX++)
                {
                    for (int loopY = 0; loopY < this.Height / rectSize + 1; loopY++)
                    {
                        if (m_chkRounded.Checked)
                        {
                            //Create rectangle to draw
                            D2D.RoundedRectangle rect = new D2D.RoundedRectangle()
                            {
                                Left = loopX * rectSize + rectSpacing,
                                Right = loopX * rectSize + rectSize - rectSpacing,
                                Top = loopY * rectSize + rectSpacing,
                                Bottom = loopY * rectSize + rectSize - rectSpacing,
                                RadiusX = (float)radius,
                                RadiusY = (float)radius
                            };

                            //Draw the rectangle (border and fill)
                            m_renderTarget.FillRoundedRectangle(m_lightBlueBrush, rect);
                            m_renderTarget.DrawRoundedRectangle(m_blackBrush, rect);
                        }
                        else
                        {
                            Rectangle rect = new Rectangle(
                                loopX * rectSize + rectSpacing,
                                loopY * rectSize + rectSpacing,
                                rectSize - rectSpacing, rectSize - rectSpacing);
                            m_renderTarget.FillRectangle(m_lightBlueBrush, rect);
                            m_renderTarget.DrawRectangle(m_blackBrush, rect);
                        }

                        //Count rectangles
                        rectangleCount++;
                    }
                }
                    stopwatch.Stop();

                    //Update output values
                    m_txtRectangleCount.Text = rectangleCount.ToString();
                    m_txtRenderDuration.Text = stopwatch.ElapsedMilliseconds.ToString();
                }
                finally
                {
                    m_renderTarget.EndDraw();
                }
            }
            else
            {
                if (m_chkHighQuality.Checked) { e.Graphics.SmoothingMode = SmoothingMode.AntiAlias; }
                else { e.Graphics.SmoothingMode = SmoothingMode.HighSpeed; }

                e.Graphics.FillRectangle(m_backBrushGdi, e.ClipRectangle);

                //Update transparency of brushes
                if (m_chkTransparent.Checked)
                {
                    m_blackBrushGdi.Color = Color.FromArgb(100, Color.Black);
                    m_lightBlueBrushGdi.Color = Color.FromArgb(100, Color.LightBlue);
                }
                else
                {
                    m_blackBrushGdi.Color = Color.Black;
                    m_lightBlueBrushGdi.Color = Color.LightBlue;
                }

                //Draw all rectangles
                int rectSize = m_scrollRectangleSize.Value;
                int rectSpacing = m_scrollRectangleSize.Value / 10;
                int radius = m_scrollRectangleSize.Value / 7;
                int rectangleCount = 0;
                Stopwatch stopwatch = new Stopwatch();
                stopwatch.Start();
                for (int loopX = 0; loopX < this.Width / rectSize + 1; loopX++)
                {
                    for (int loopY = 0; loopY < this.Height / rectSize + 1; loopY++)
                    {
                        Rectangle rect = new Rectangle(
                            loopX * rectSize + rectSpacing,
                            loopY * rectSize + rectSpacing,
                            rectSize - rectSpacing, rectSize - rectSpacing);
                        e.Graphics.FillRectangle(m_lightBlueBrushGdi, rect);
                        e.Graphics.DrawRectangle(m_blackBrushGdi, rect);

                        //Count rectangles
                        rectangleCount++;
                    }
                }
                stopwatch.Stop();

                //Update output values
                m_txtRectangleCount.Text = rectangleCount.ToString();
                m_txtRenderDuration.Text = stopwatch.ElapsedMilliseconds.ToString();
            }
        }

        /// <summary>
        /// Is debug mode enabled?
        /// </summary>
        [Category("Direct2D")]
        public bool IsDebugEnabled
        {
            get { return m_debugMode; }
            set { m_debugMode = value; }
        }
    }
}
