﻿using System;
using System.ComponentModel;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using SlimDX;

//Some namespace mappings
using D2D    = SlimDX.Direct2D;
using DWrite = SlimDX.DirectWrite;

namespace Articles.Tutorials.Direct2D
{
    public class Chapter3 : UserControl
    {
        public const string TEXT_TOP_LEFT = "This is a simple text rendered using RenderTarget.DrawText. No TextLayout was generated here.";
        public const string TEXT_TOP_RIGHT = "This is a simple text rendered using RenderTarget.DrawText. Textlayout was generated with size (400,100).";

        //Resources for Direct2D rendering
        private D2D.Factory m_factory;
        private D2D.WindowRenderTarget m_renderTarget;
        private D2D.LinearGradientBrush m_backBrushEx;
        private D2D.GradientStopCollection m_backBrushGradient;
        private D2D.SolidColorBrush m_blackBrush;

        //Resources for DirectWrite rendering
        private DWrite.Factory m_factoryWrite;
        private DWrite.TextFormat m_textFormatCentered;
        private DWrite.TextFormat m_textFormatFlow;
        private DWrite.TextLayout m_textLayoutFlow;

        //Some generic members
        private Brush m_backBrushGdi;
        private bool m_initialized;
        private bool m_debugMode;

        /// <summary>
        /// Initializes a new instance of the <see cref="Direct2DPanel"/> class.
        /// </summary>
        public Chapter3()
        {
            //Update control styles
            this.SetStyle(ControlStyles.AllPaintingInWmPaint, true);
            this.SetStyle(ControlStyles.Opaque, true);
            this.SetStyle(ControlStyles.ResizeRedraw, true);

            //Update back brush
            m_backBrushGdi = new SolidBrush(this.BackColor);
        }

        /// <summary>
        /// Loads all graphics resources.
        /// </summary>
        private void InitializeGraphics()
        {
            //Get requested debug level
            D2D.DebugLevel debugLevel = D2D.DebugLevel.None;
            if (m_debugMode) { debugLevel = D2D.DebugLevel.Error; }

            //Create factory objects
            m_factory = new D2D.Factory(D2D.FactoryType.SingleThreaded, debugLevel);
            m_factoryWrite = new DWrite.Factory(DWrite.FactoryType.Shared);

            //Create the render target
            m_renderTarget = new D2D.WindowRenderTarget(m_factory, new D2D.WindowRenderTargetProperties()
            {
                Handle = this.Handle,
                PixelSize = this.Size,
                PresentOptions = D2D.PresentOptions.Immediately
            });

            //Create linear gradient brush
            D2D.GradientStop[] gradientStops = new D2D.GradientStop[]
            {
                new D2D.GradientStop(){ Position = 0f, Color = new Color4(Color.LightGray) },
                new D2D.GradientStop(){ Position = 1f, Color = new Color4(Color.LightSteelBlue) }
            };
            m_backBrushGradient = new D2D.GradientStopCollection(m_renderTarget, gradientStops);
            m_backBrushEx = new D2D.LinearGradientBrush(
                m_renderTarget, 
                m_backBrushGradient, 
                new D2D.LinearGradientBrushProperties()
            {
                StartPoint = new PointF(0, this.Height),
                EndPoint = new PointF(0, 0)
            });

            //Create solid brush
            m_blackBrush = new D2D.SolidColorBrush(m_renderTarget, new Color4(Color.Black));

            //Create DirectWrite resources
            m_textFormatCentered = m_factoryWrite.CreateTextFormat(
                "Gabriola", 
                DWrite.FontWeight.Normal, DWrite.FontStyle.Normal, DWrite.FontStretch.Normal, 
                72f, "de-de");
            m_textFormatCentered.TextAlignment = DWrite.TextAlignment.Center;
            m_textFormatCentered.ParagraphAlignment = DWrite.ParagraphAlignment.Center;

            m_textFormatFlow = m_factoryWrite.CreateTextFormat(
                "Arial",
                DWrite.FontWeight.Normal, DWrite.FontStyle.Normal, DWrite.FontStretch.Normal,
                20f, "de-de");
            m_textFormatFlow.TextAlignment = DWrite.TextAlignment.Leading;
            m_textFormatFlow.ParagraphAlignment = DWrite.ParagraphAlignment.Near;

            m_textLayoutFlow = m_factoryWrite.CreateTextLayout(
                TEXT_TOP_RIGHT, m_textFormatFlow,
                400f, 100f);

            //Update initialization flag
            m_initialized = true;
        }

        /// <summary>
        /// Unloads all graphics resources.
        /// </summary>
        private void UnloadGraphics()
        {
            if (m_textLayoutFlow != null) { m_textLayoutFlow.Dispose(); }
            if (m_textFormatFlow != null) { m_textFormatFlow.Dispose(); }
            if (m_textFormatCentered != null) { m_textFormatCentered.Dispose(); }
            if (m_blackBrush != null) { m_blackBrush.Dispose(); }
            if (m_backBrushEx != null) { m_backBrushEx.Dispose(); }
            if (m_backBrushGradient != null) { m_backBrushGradient.Dispose(); }
            if (m_renderTarget != null) { m_renderTarget.Dispose(); }
            if (m_factoryWrite != null) { m_factoryWrite.Dispose(); }
            if (m_factory != null) { m_factory.Dispose(); }

            m_backBrushEx = null;
            m_backBrushGradient = null;
            m_renderTarget = null;
            m_factory = null;
        }

        /// <summary>
        /// Called when window opens.
        /// </summary>
        /// <param name="e">An <see cref="T:System.EventArgs"/> that contains the event data.</param>
        protected override void OnHandleCreated(EventArgs e)
        {
            base.OnHandleCreated(e);

            if ((!m_initialized) || (!this.DesignMode))
            {
                InitializeGraphics();
            }
        }

        /// <summary>
        /// Called when window closes.
        /// </summary>
        /// <param name="e">An <see cref="T:System.EventArgs"/> that contains the event data.</param>
        protected override void OnHandleDestroyed(EventArgs e)
        {
            base.OnHandleDestroyed(e);

            if (m_initialized)
            {
                UnloadGraphics();
            }
        }

        /// <summary>
        /// Called when background color has changed.
        /// </summary>
        /// <param name="e">An <see cref="T:System.EventArgs"/> that contains the event data.</param>
        protected override void OnBackColorChanged(EventArgs e)
        {
            base.OnBackColorChanged(e);

            //Update back brush
            if (m_backBrushGdi != null) { m_backBrushGdi.Dispose(); }
            m_backBrushGdi = new SolidBrush(this.BackColor);
        }

        /// <summary>
        /// Called when size of the control changed.
        /// </summary>
        /// <param name="e">An <see cref="T:System.EventArgs"/> that contains the event data.</param>
        protected override void OnResize(EventArgs e)
        {
            base.OnResize(e);

            if ((this.Width > 0) && (this.Height > 0))
            {
                if (m_initialized)
                {
                    //Resize render target
                    m_renderTarget.Resize(this.Size);
                    m_backBrushEx.StartPoint = new PointF(0, this.Height);
                }
            }
        }

        /// <summary>
        /// Called when control has to paint itself.
        /// </summary>
        /// <param name="e">A <see cref="T:System.Windows.Forms.PaintEventArgs"/> that contains the event data.</param>
        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);

            if (m_initialized)
            {
                m_renderTarget.BeginDraw();
                try
                {
                    m_renderTarget.Clear(new Color4(this.BackColor));

                    //Perform all Direct2D rendering here
                    m_renderTarget.FillRectangle(
                        m_backBrushEx,
                        new Rectangle(new Point(), this.Size));

                    //Draw simple text
                    m_renderTarget.DrawText(
                        "Hello World using DirectWrite!", m_textFormatCentered,
                        new Rectangle(new Point(), this.Size), m_blackBrush);

                    //Render top left rectangle
                    {
                        Rectangle textRect = new Rectangle(
                            10, 10,
                            this.Width / 2 - 20, this.Height / 3 - 20);

                        m_renderTarget.DrawText(
                            TEXT_TOP_LEFT,
                            m_textFormatFlow,
                            textRect, m_blackBrush,
                            D2D.DrawTextOptions.Clip, DWrite.MeasuringMethod.Natural);
                    }

                    //Render top right rectangle
                    {
                        Point topRight = new Point(this.Width / 2 + 10, 10);
                        m_renderTarget.DrawTextLayout(topRight, m_textLayoutFlow, m_blackBrush, D2D.DrawTextOptions.Clip);
                    }
                }
                finally
                {
                    m_renderTarget.EndDraw();
                }
            }
            else
            {
                e.Graphics.FillRectangle(m_backBrushGdi, e.ClipRectangle);
            }
        }

        /// <summary>
        /// Is debug mode enabled?
        /// </summary>
        [Category("Direct2D")]
        public bool IsDebugEnabled
        {
            get { return m_debugMode; }
            set { m_debugMode = value; }
        }
    }
}
