﻿using System;
using System.ComponentModel;
using System.Diagnostics;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using SlimDX;

//Some namespace mappings
using D2D  = SlimDX.Direct2D;
using DXGI = SlimDX.DXGI;

namespace Articles.Tutorials.Direct2D
{
    public partial class Chapter4 : UserControl
    {
        //Resources for Direct2D rendering
        private D2D.Factory m_factory;
        private D2D.WindowRenderTarget m_renderTarget;
        private D2D.LinearGradientBrush m_backBrushEx;
        private D2D.GradientStopCollection m_backBrushGradient;
        private D2D.Bitmap m_puzzleBitmap;

        //Some generic members
        private Bitmap m_puzzleBitmapGdi;
        private Brush m_backBrushGdi;
        private bool m_initialized;
        private bool m_debugMode;

        /// <summary>
        /// Initializes a new instance of the <see cref="Direct2DPanel"/> class.
        /// </summary>
        public Chapter4()
        {
            this.InitializeComponent();

            //Update control styles
            this.SetStyle(ControlStyles.AllPaintingInWmPaint, true);
            this.SetStyle(ControlStyles.Opaque, true);
            this.SetStyle(ControlStyles.ResizeRedraw, true);

            //Update back brush
            m_backBrushGdi = new SolidBrush(this.BackColor);
        }

        /// <summary>
        /// Loads a Direct2D bitmap from the given file.
        /// </summary>
        /// <param name="file">Source file name.</param>
        public D2D.Bitmap LoadBitmap(string file)
        {
            D2D.Bitmap result = null;

            //Create Gdi Bitmap 
            Bitmap drawingBitmap = Bitmap.FromFile(file) as Bitmap;

            //Load the Direct2D resource from the gdi resource
            result = LoadBitmap(drawingBitmap);

            //Dispose the gdi bitmap after loading
            drawingBitmap.Dispose();

            return result;
        }

        /// <summary>
        /// Load a Direct2D bitmap from the given gdi resource.
        /// </summary>
        /// <param name="drawingBitmap">The gdi resource.</param>
        public D2D.Bitmap LoadBitmap(Bitmap drawingBitmap)
        {
            D2D.Bitmap result = null;

            //Lock the gdi resource
            BitmapData drawingBitmapData = drawingBitmap.LockBits(
                new Rectangle(0, 0, drawingBitmap.Width, drawingBitmap.Height),
                ImageLockMode.ReadOnly, PixelFormat.Format32bppPArgb);

            //Prepare loading the image from gdi resource
            DataStream dataStream = new DataStream(
                drawingBitmapData.Scan0,
                drawingBitmapData.Stride * drawingBitmapData.Height,
                true, false);
            D2D.BitmapProperties properties = new D2D.BitmapProperties();
            properties.PixelFormat = new D2D.PixelFormat(
                DXGI.Format.B8G8R8A8_UNorm,
                D2D.AlphaMode.Premultiplied);

            //Load the image from the gdi resource
            result = new D2D.Bitmap(
                m_renderTarget,
                new Size(drawingBitmap.Width, drawingBitmap.Height),
                dataStream, drawingBitmapData.Stride,
                properties);

            //Unlock the gdi resource
            drawingBitmap.UnlockBits(drawingBitmapData);

            return result;
        }

        /// <summary>
        /// Loads all graphics resources.
        /// </summary>
        private void InitializeGraphics()
        {
            //Get requested debug level
            D2D.DebugLevel debugLevel = D2D.DebugLevel.None;
            if (m_debugMode) { debugLevel = D2D.DebugLevel.Error; }

            //Create factory object
            m_factory = new D2D.Factory(D2D.FactoryType.SingleThreaded, debugLevel);

            //Create the render target
            m_renderTarget = new D2D.WindowRenderTarget(m_factory, new D2D.WindowRenderTargetProperties()
            {
                Handle = this.Handle,
                PixelSize = this.Size,
                PresentOptions = D2D.PresentOptions.Immediately
            });

            //Create linear gradient brush
            D2D.GradientStop[] gradientStops = new D2D.GradientStop[]
            {
                new D2D.GradientStop(){ Position = 0f, Color = new Color4(Color.LightGray) },
                new D2D.GradientStop(){ Position = 1f, Color = new Color4(Color.LightSteelBlue) }
            };
            m_backBrushGradient = new D2D.GradientStopCollection(m_renderTarget, gradientStops);
            m_backBrushEx = new D2D.LinearGradientBrush(
                m_renderTarget, 
                m_backBrushGradient, 
                new D2D.LinearGradientBrushProperties()
            {
                StartPoint = new PointF(0, this.Height),
                EndPoint = new PointF(0, 0)
            });

            //Create gdi backbrush
            if (m_backBrushGdi != null) { m_backBrushGdi.Dispose(); }
            m_backBrushGdi = new LinearGradientBrush(
                new Point(0, this.Height),
                new Point(0, 0),
                Color.LightGray,
                Color.LightSteelBlue);

            //Load the bitmap
            m_puzzleBitmapGdi = Properties.Resources.Puzzle;
            m_puzzleBitmap = LoadBitmap(m_puzzleBitmapGdi);

            //Update initialization flag
            m_initialized = true;
        }

        /// <summary>
        /// Unloads all graphics resources.
        /// </summary>
        private void UnloadGraphics()
        {
            if (m_puzzleBitmap != null) { m_puzzleBitmap.Dispose(); }
            if (m_backBrushEx != null) { m_backBrushEx.Dispose(); }
            if (m_backBrushGradient != null) { m_backBrushGradient.Dispose(); }
            if (m_renderTarget != null) { m_renderTarget.Dispose(); }
            if (m_factory != null) { m_factory.Dispose(); }

            m_backBrushEx = null;
            m_backBrushGradient = null;
            m_renderTarget = null;
            m_factory = null;
        }

        /// <summary>
        /// Called when window opens.
        /// </summary>
        /// <param name="e">An <see cref="T:System.EventArgs"/> that contains the event data.</param>
        protected override void OnHandleCreated(EventArgs e)
        {
            base.OnHandleCreated(e);

            if ((!m_initialized) || (!this.DesignMode))
            {
                InitializeGraphics();
            }
        }

        /// <summary>
        /// Called when window closes.
        /// </summary>
        /// <param name="e">An <see cref="T:System.EventArgs"/> that contains the event data.</param>
        protected override void OnHandleDestroyed(EventArgs e)
        {
            base.OnHandleDestroyed(e);

            if (m_initialized)
            {
                UnloadGraphics();
            }
        }

        /// <summary>
        /// Called when background color has changed.
        /// </summary>
        /// <param name="e">An <see cref="T:System.EventArgs"/> that contains the event data.</param>
        protected override void OnBackColorChanged(EventArgs e)
        {
            base.OnBackColorChanged(e);

            //Update back brush
            if (m_backBrushGdi != null) { m_backBrushGdi.Dispose(); }
            m_backBrushGdi = new SolidBrush(this.BackColor);
        }

        /// <summary>
        /// Called when user changes a parameter.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void OnChkParameterCheckedChanged(object sender, EventArgs e)
        {
            this.Invalidate();
        }

        /// <summary>
        /// Called when user changes the renderer.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void OnRadGdiCheckedChanged(object sender, EventArgs e)
        {
            if (m_radGdi.Checked)
            {
                m_chkTransparent.Enabled = false;
                m_chkTransparent.Checked = false;
                this.DoubleBuffered = true;
                this.Invalidate();
            }
        }

        /// <summary>
        /// Called when user changes the renderer.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void OnRadD2dCheckedChanged(object sender, EventArgs e)
        {
            if (m_radD2d.Checked)
            {
                m_chkTransparent.Enabled = true;
                this.DoubleBuffered = false;
                this.Invalidate();
            }
        }

        /// <summary>
        /// Called when size of the control changed.
        /// </summary>
        /// <param name="e">An <see cref="T:System.EventArgs"/> that contains the event data.</param>
        protected override void OnResize(EventArgs e)
        {
            base.OnResize(e);

            if ((this.Width > 0) && (this.Height > 0))
            {
                if (m_backBrushGdi != null) { m_backBrushGdi.Dispose(); }
                m_backBrushGdi = new LinearGradientBrush(
                    new Point(0, this.Height),
                    new Point(0, 0),
                    Color.LightGray,
                    Color.LightSteelBlue);

                if (m_initialized)
                {
                    //Resize render target
                    m_renderTarget.Resize(this.Size);
                    m_backBrushEx.StartPoint = new PointF(0, this.Height);
                }
            }
        }

        /// <summary>
        /// Called when uswer changes the size of rectangles.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="System.Windows.Forms.ScrollEventArgs"/> instance containing the event data.</param>
        private void OnScrollRectangleSizeScroll(object sender, ScrollEventArgs e)
        {
            this.Invalidate();
        }

        /// <summary>
        /// Called when control has to paint itself.
        /// </summary>
        /// <param name="e">A <see cref="T:System.Windows.Forms.PaintEventArgs"/> that contains the event data.</param>
        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);

            if (m_initialized && m_radD2d.Checked)
            {
                m_renderTarget.BeginDraw();
                try
                {
                    m_renderTarget.Clear(new Color4(this.BackColor));

                    //Perform all Direct2D rendering here
                    m_renderTarget.FillRectangle(
                        m_backBrushEx,
                        new Rectangle(new Point(), this.Size));

                    //Apply quality
                    D2D.InterpolationMode interpolationMode = D2D.InterpolationMode.NearestNeighbor;
                    if (m_chkHighQuality.Checked) { interpolationMode = D2D.InterpolationMode.Linear; }

                    //Draw all images
                    int imageCount = 0;
                    float transparency = m_chkTransparent.Checked ? 0.4f : 1f;
                    int imageWidth = m_scrollRectangleSize.Value;
                    int imageHeight = m_scrollRectangleSize.Value;
                    Stopwatch stopwatch = new Stopwatch();
                    stopwatch.Start();
                    for (int loopX = 0; loopX < this.Width / imageWidth + 1; loopX++)
                    {
                        for (int loopY = 0; loopY < this.Height / imageHeight + 1; loopY++)
                        {
                            //Draw the bitmap
                            m_renderTarget.DrawBitmap(
                                m_puzzleBitmap,
                                new RectangleF(
                                    loopX * imageWidth, loopY * imageHeight, 
                                    imageWidth, imageHeight),
                                transparency,
                                interpolationMode);
                            imageCount++;
                        }
                    }
                    stopwatch.Stop();

                    //Update output values
                    m_txtRectangleCount.Text = imageCount.ToString();
                    m_txtRenderDuration.Text = stopwatch.ElapsedMilliseconds.ToString();
                }
                finally
                {
                    m_renderTarget.EndDraw();
                }
            }
            else
            {
                //Switch quality mode
                if (m_chkHighQuality.Checked)
                {
                    e.Graphics.SmoothingMode = SmoothingMode.AntiAlias;
                    e.Graphics.InterpolationMode = InterpolationMode.High;
                }
                else 
                { 
                    e.Graphics.SmoothingMode = SmoothingMode.HighSpeed;
                    e.Graphics.InterpolationMode = InterpolationMode.Low;
                }

                //Draw the background
                e.Graphics.FillRectangle(m_backBrushGdi, e.ClipRectangle);

                //Draw all images
                int imageCount = 0;
                float transparency = m_chkTransparent.Checked ? 0.4f : 1f;
                int imageWidth = m_scrollRectangleSize.Value;
                int imageHeight = m_scrollRectangleSize.Value;
                Stopwatch stopwatch = new Stopwatch();
                stopwatch.Start();
                for (int loopX = 0; loopX < this.Width / imageWidth + 1; loopX++)
                {
                    for (int loopY = 0; loopY < this.Height / imageHeight + 1; loopY++)
                    {
                        //Draw the bitmap
                        e.Graphics.DrawImage(
                            m_puzzleBitmapGdi,
                            new RectangleF(
                                loopX * imageWidth, loopY * imageHeight,
                                imageWidth, imageHeight));
                        imageCount++;
                    }
                }
                stopwatch.Stop();

                //Update output values
                m_txtRectangleCount.Text = imageCount.ToString();
                m_txtRenderDuration.Text = stopwatch.ElapsedMilliseconds.ToString();
            }
        }

        /// <summary>
        /// Is debug mode enabled?
        /// </summary>
        [Category("Direct2D")]
        public bool IsDebugEnabled
        {
            get { return m_debugMode; }
            set { m_debugMode = value; }
        }
    }
}
